/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#include "config.h"

#include "gtkmenuprivate.h"
#include "gtkmenuitemprivate.h"
#include "gtktearoffmenuitem.h"
#include "gtkintl.h"

#define ARROW_SIZE 10
#define TEAR_LENGTH 5
#define BORDER_SPACING  3

struct _GtkTearoffMenuItemPrivate
{
  guint torn_off : 1;
};

static void gtk_tearoff_menu_item_get_preferred_width  (GtkWidget      *widget,
							gint           *minimum,
							gint           *natural);
static void gtk_tearoff_menu_item_get_preferred_height (GtkWidget      *widget,
							gint           *minimum,
							gint           *natural);
static gboolean gtk_tearoff_menu_item_draw             (GtkWidget      *widget,
							cairo_t        *cr);
static void gtk_tearoff_menu_item_activate             (GtkMenuItem    *menu_item);
static void gtk_tearoff_menu_item_parent_set           (GtkWidget      *widget,
							GtkWidget      *previous);

G_DEFINE_TYPE (GtkTearoffMenuItem, gtk_tearoff_menu_item, GTK_TYPE_MENU_ITEM)

GtkWidget*
gtk_tearoff_menu_item_new (void)
{
  return g_object_new (GTK_TYPE_TEAROFF_MENU_ITEM, NULL);
}

static void
gtk_tearoff_menu_item_class_init (GtkTearoffMenuItemClass *klass)
{
  GtkWidgetClass *widget_class;
  GtkMenuItemClass *menu_item_class;

  widget_class = (GtkWidgetClass*) klass;
  menu_item_class = (GtkMenuItemClass*) klass;

  widget_class->draw = gtk_tearoff_menu_item_draw;
  widget_class->get_preferred_width = gtk_tearoff_menu_item_get_preferred_width;
  widget_class->get_preferred_height = gtk_tearoff_menu_item_get_preferred_height;
  widget_class->parent_set = gtk_tearoff_menu_item_parent_set;

  menu_item_class->activate = gtk_tearoff_menu_item_activate;

  g_type_class_add_private (klass, sizeof (GtkTearoffMenuItemPrivate));
}

static void
gtk_tearoff_menu_item_init (GtkTearoffMenuItem *tearoff_menu_item)
{
  GtkTearoffMenuItemPrivate *priv;

  tearoff_menu_item->priv = G_TYPE_INSTANCE_GET_PRIVATE (tearoff_menu_item,
                                                         GTK_TYPE_TEAROFF_MENU_ITEM,
                                                         GtkTearoffMenuItemPrivate);
  priv = tearoff_menu_item->priv;

  priv->torn_off = FALSE;
}

static void
gtk_tearoff_menu_item_get_preferred_width (GtkWidget      *widget,
					   gint           *minimum,
					   gint           *natural)
{
  GtkStyle *style;
  guint border_width;

  style = gtk_widget_get_style (widget);

  border_width = gtk_container_get_border_width (GTK_CONTAINER (widget));
  *minimum = *natural = (border_width + style->xthickness + BORDER_SPACING) * 2;
}

static void
gtk_tearoff_menu_item_get_preferred_height (GtkWidget      *widget,
					    gint           *minimum,
					    gint           *natural)
{
  GtkStyle *style;
  GtkWidget *parent;
  guint border_width;

  style = gtk_widget_get_style (widget);

  border_width = gtk_container_get_border_width (GTK_CONTAINER (widget));
  *minimum = *natural = (border_width + style->ythickness) * 2;

  parent = gtk_widget_get_parent (widget);
  if (GTK_IS_MENU (parent) && GTK_MENU (parent)->priv->torn_off)
    {
      *minimum += ARROW_SIZE;
      *natural += ARROW_SIZE;
    }
  else
    {
      *minimum += style->ythickness + 4;
      *natural += style->ythickness + 4;
    }
}

static gboolean
gtk_tearoff_menu_item_draw (GtkWidget *widget,
                            cairo_t   *cr)
{
  GtkMenuItem *menu_item;
  GtkShadowType shadow_type;
  GtkStateType state;
  GtkStyle *style;
  gint x, y, width, height;
  gint right_max;
  guint border_width;
  GtkArrowType arrow_type;
  GtkTextDirection direction;
  GtkWidget *parent;
  GdkWindow *window;

  menu_item = GTK_MENU_ITEM (widget);
  style = gtk_widget_get_style (widget);
  window = gtk_widget_get_window (widget);
  direction = gtk_widget_get_direction (widget);
  state = gtk_widget_get_state (widget);

  border_width = gtk_container_get_border_width (GTK_CONTAINER (menu_item));
  x = border_width;
  y = border_width;
  width = gtk_widget_get_allocated_width (widget) - border_width * 2;
  height = gtk_widget_get_allocated_height (widget) - border_width * 2;
  right_max = x + width;

  if (state == GTK_STATE_PRELIGHT)
    {
      gint selected_shadow_type;
      
      gtk_widget_style_get (widget,
                            "selected-shadow-type", &selected_shadow_type,
                            NULL);
      gtk_paint_box (style,
                     cr,
                     GTK_STATE_PRELIGHT,
                     selected_shadow_type,
                     widget, "menuitem",
                     x, y, width, height);
    }

  parent = gtk_widget_get_parent (widget);
  if (GTK_IS_MENU (parent) && GTK_MENU (parent)->priv->torn_off)
    {
      gint arrow_x;

      if (state == GTK_STATE_PRELIGHT)
        shadow_type = GTK_SHADOW_IN;
      else
        shadow_type = GTK_SHADOW_OUT;

      if (menu_item->priv->toggle_size > ARROW_SIZE)
        {
          if (direction == GTK_TEXT_DIR_LTR) {
            arrow_x = x + (menu_item->priv->toggle_size - ARROW_SIZE)/2;
            arrow_type = GTK_ARROW_LEFT;
          }
          else {
            arrow_x = x + width - menu_item->priv->toggle_size + (menu_item->priv->toggle_size - ARROW_SIZE)/2;
            arrow_type = GTK_ARROW_RIGHT;
          }
          x += menu_item->priv->toggle_size + BORDER_SPACING;
        }
      else
        {
          if (direction == GTK_TEXT_DIR_LTR)
            {
              arrow_x = ARROW_SIZE / 2;
              arrow_type = GTK_ARROW_LEFT;
            }
          else
            {
              arrow_x = x + width - 2 * ARROW_SIZE + ARROW_SIZE / 2;
              arrow_type = GTK_ARROW_RIGHT;
            }
          x += 2 * ARROW_SIZE;
        }


      gtk_paint_arrow (style, cr,
                       state, shadow_type,
                       widget, "tearoffmenuitem",
                       arrow_type, FALSE,
                       arrow_x, y + height / 2 - 5,
                       ARROW_SIZE, ARROW_SIZE);
    }

  while (x < right_max)
    {
      gint x1, x2;

      if (direction == GTK_TEXT_DIR_LTR) {
        x1 = x;
        x2 = MIN (x + TEAR_LENGTH, right_max);
      }
      else {
        x1 = right_max - x;
        x2 = MAX (right_max - x - TEAR_LENGTH, 0);
      }

      gtk_paint_hline (style, cr, GTK_STATE_NORMAL,
                       widget, "tearoffmenuitem",
                       x1, x2, y + (height - style->ythickness) / 2);
      x += 2 * TEAR_LENGTH;
    }

  return FALSE;
}

static void
gtk_tearoff_menu_item_activate (GtkMenuItem *menu_item)
{
  GtkWidget *parent;

  parent = gtk_widget_get_parent (GTK_WIDGET (menu_item));
  if (GTK_IS_MENU (parent))
    {
      GtkMenu *menu = GTK_MENU (parent);

      gtk_widget_queue_resize (GTK_WIDGET (menu_item));
      gtk_menu_set_tearoff_state (GTK_MENU (parent),
				  !menu->priv->torn_off);
    }
}

static void
tearoff_state_changed (GtkMenu            *menu,
		       GParamSpec         *pspec,
		       gpointer            data)
{
  GtkTearoffMenuItem *tearoff_menu_item = GTK_TEAROFF_MENU_ITEM (data);
  GtkTearoffMenuItemPrivate *priv = tearoff_menu_item->priv;

  priv->torn_off = gtk_menu_get_tearoff_state (menu);
}

static void
gtk_tearoff_menu_item_parent_set (GtkWidget *widget,
				  GtkWidget *previous)
{
  GtkTearoffMenuItem *tearoff_menu_item = GTK_TEAROFF_MENU_ITEM (widget);
  GtkTearoffMenuItemPrivate *priv = tearoff_menu_item->priv;
  GtkMenu *menu;
  GtkWidget *parent;

  parent = gtk_widget_get_parent (widget);
  menu = GTK_IS_MENU (parent) ? GTK_MENU (parent) : NULL;

  if (previous)
    g_signal_handlers_disconnect_by_func (previous, 
					  tearoff_state_changed, 
					  tearoff_menu_item);
  
  if (menu)
    {
      priv->torn_off = gtk_menu_get_tearoff_state (menu);
      g_signal_connect (menu, "notify::tearoff-state", 
			G_CALLBACK (tearoff_state_changed), 
			tearoff_menu_item);
    }  
}
